//-----------------------------------------------------------------------------
// Copyright (c) 2007 dhpoware. All Rights Reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.
//-----------------------------------------------------------------------------

#include "thread.h"

Thread::Thread()
{
    m_hThread = 0;
    m_threadId = 0;
    m_pszThreadName = 0;
    m_pRunnable = 0;
    m_pfnThreadFunc = 0;
}

Thread::~Thread()
{
    if (m_hThread)
    {
        CloseHandle(m_hThread);
        m_hThread = 0;
    }
}

bool Thread::create(unsigned int stackSize)
{
    m_hThread = reinterpret_cast<HANDLE>(_beginthreadex( 0, stackSize,
                    threadFunc, this, CREATE_SUSPENDED, &m_threadId));

    if (m_hThread)
        return true;

    return false;
}

bool Thread::create(const char *pszThreadName, unsigned int stackSize)
{
    m_pszThreadName = pszThreadName;
    return create(stackSize);
}

unsigned int Thread::id() const
{
    return m_threadId;
}

const char *Thread::name() const
{
    return m_pszThreadName;
}

void Thread::start()
{
    ResumeThread(m_hThread);
}

void Thread::start(void (*pfnThreadFunc)())
{
    m_pfnThreadFunc = pfnThreadFunc;
    ResumeThread(m_hThread);
}

void Thread::start(IRunnable *pRunnable)
{
    m_pRunnable = pRunnable;
    ResumeThread(m_hThread);
}

void Thread::join()
{
    WaitForSingleObject(m_hThread, INFINITE);
}

Thread::Priority Thread::getPriority() const
{
    switch (GetThreadPriority(m_hThread))
    {
    case THREAD_PRIORITY_ABOVE_NORMAL:
        return PRIORITY_ABOVE_NORMAL;

    case THREAD_PRIORITY_BELOW_NORMAL:
        return PRIORITY_BELOW_NORMAL;

    case THREAD_PRIORITY_HIGHEST:
        return PRIORITY_HIGHEST;

    case THREAD_PRIORITY_LOWEST:
        return PRIORITY_LOWEST;
    
    case THREAD_PRIORITY_NORMAL:   
    default:
        return PRIORITY_NORMAL;
    }
}

void Thread::setPriority(Thread::Priority priority)
{
    int nPriority = 0;

    switch (priority)
    {
    case PRIORITY_LOWEST:
        nPriority = THREAD_PRIORITY_LOWEST;
        break;

    case PRIORITY_BELOW_NORMAL:
        nPriority = THREAD_PRIORITY_BELOW_NORMAL;
        break;

    case PRIORITY_ABOVE_NORMAL:
        nPriority = THREAD_PRIORITY_ABOVE_NORMAL;
        break;

    case PRIORITY_HIGHEST:
        nPriority = THREAD_PRIORITY_HIGHEST;
        break;
    
    case PRIORITY_NORMAL:
    default:
        nPriority = THREAD_PRIORITY_NORMAL;
        break;
    }

    SetThreadPriority(m_hThread, nPriority);
}

void Thread::run()
{
}

unsigned int __stdcall Thread::threadFunc(void *pArgs)
{
    if (Thread *pThread = reinterpret_cast<Thread *>(pArgs))
    {
        pThread->setThreadName();

        if (pThread->m_pfnThreadFunc)
            pThread->m_pfnThreadFunc();
        else if (pThread->m_pRunnable)
            pThread->m_pRunnable->run();
        else
            pThread->run();
    }

    _endthreadex(0);
    return 0;
}

void Thread::setThreadName()
{
    // Based on code appearing in the technical article by Bruce Dawson titled
    // "Coding For Multiple Cores on Xbox 360 and Microsoft Windows" included
    // with the December 2006 DirectX SDK update.

    if (!m_pszThreadName)
        return;

    struct ThreadNameInfo
    {
        DWORD dwType;       // must be 0x1000
        LPCSTR pszName;     // pointer to name (in user address space)
        DWORD dwThreadId;   // thread ID (-1 = caller thread)
        DWORD dwFlags;      // reserved for future use; must be zero
    };

    ThreadNameInfo info = {0x1000, m_pszThreadName, m_threadId, 0};

    __try
    {
        RaiseException(0x406D1388, 0, sizeof(info) / sizeof(DWORD),
            reinterpret_cast<DWORD*>(&info));
    }
    __except (EXCEPTION_CONTINUE_EXECUTION)
    {
    }
}